-- D42 dumper V1.1

strUsage = [[
Usage: lua d42_dump.lua <data dir>
Create a textual dump of a d42 adventure.
The data dir must contain all files of the game.
]]

--MODE_HOTSPOT = true
if not MODE_HOTSPOT then
	-- for standard games 
	OJ_OBJ_OFFSET = 0xe03
else
	-- for Hotspot/32
	OJ_OBJ_OFFSET = 3
end

function _G.printf(...) print(string.format(...)) end
_G.verbosity = 0
function _G.vbsprintf(lv, ...) if verbosity >= lv then printf(...) end end

function loadFile(name)
	local file, msg = io.open(name, "rb")
	if not file then return nil end
	assert(file, msg)
	local data = file:read("*a")
	file:close()
	assert(data, "error reading file " .. name)
	vbsprintf(2, "file %s - %d bytes read", name, data:len())
	return data
end

function writeFile(name, data)
	local file, msg = io.open(name, "wb+")
	assert(file, msg)
	file:write(data)
	file:close()	
	vbsprintf(2, "file %s written - %d bytes", name, data:len())
end

function depackD4(bin)
	local pos = 0
	local function getbyte() 
		pos = pos + 1
		return bin:byte(pos)
	end
	
	local unpacked = {}
	local function putbyte(b)
		table.insert(unpacked, string.char(b))
	end
	
	putbyte(getbyte()) -- copy load addr
	putbyte(getbyte())
	
	local packbyte = getbyte()
	
	while (pos<bin:len()) do
		local b = getbyte()
		if b == packbyte then
			local count = getbyte()
			b = getbyte()
			if count== 0 then break end
			for i=1,count do putbyte(b) end
		else
			putbyte(b)
		end
	end

	if #unpacked < 0x1f02 then
		printf("unpacked room file is too short: 0x%x", #unpacked)
	elseif #unpacked > 0x2002 then
		printf("unpacked room file exceeds area: 0x%x", #unpacked)
	end
		
	if pos < bin:len() then
		printf("depackD4: End marker before EOF: pos = 0x%x, packed len = 0x%x, unpacked len = 0x%x",
		pos, bin:len(), #unpacked)
	end
	
	return table.concat(unpacked)
end


bit = {}
function bit.band(a, b)
	local x=128
	local c=0
	while x>=1 do
		if a>=x and b>=x then
			c=c+x
		end
		a = a % x
		b = b % x
		x = x/2
	end
	return c
end


-- Allgemeindaten:
-- OJ/O2 Liste der Objekte
-- FL    Liste der Flags
-- NW    Objektverteilung auf Rume
-- BA    Allgemeinbedingungen
--       Allgemeintexte
--       
-- Raumdaten:
-- Bedingungen
-- Texte
-- Richtungsflags
-- Musiknr.

-- BA
-- Adresse Inhalt
-- a000-a3ff 64 Bedingungen
-- a400-a4ef Allgemeintext 1
-- ... ...
-- b300-b3ef Allgemeintext 16

-- OJ/O2
-- Objekte $40-$7f (File o2)
-- Adresse Inhalt
-- 5000-53ff Deskriptoren fr Objekte $40-$7f
-- Objekte $00-$37 (File oj)
-- 6400-71ff Sprites fr Objekte $00-$37
-- 643f, 647f Sprite-Multicolorfarben
-- 7200-757f Deskriptoren fr Objekte $00-$37
-- 7580-75ff frei


-- NW
-- Adresse Funktion
-- 0c00-0c7f Objekt 0 fr die Raume 00-7f
-- ... ...
-- 0f80-0?f Objekt 7 fr die Raume 00-7f

-- Objektdescriptor:
-- Byte (Offset) Funktion
-- 0-14 Objektname, Screencodes aufgefllt mit Leerzeichen
-- 15 Bit 0 0: Objekt kann genommen werden, 1: unbewegliches Objekt
-- 15 Bit 1 0: Gegenstand, 1: Person
-- 15 Bit 2 0: Hires-Sprites, 1: Multicolor-Sprite
-- 15 Bit 3 reserviert
-- 15 Bits 4-7 Spritefarbe

-- Textcodierung:
-- Hexcode Zeichen
-- $00 oder $20 Space
-- $1e Text-Endemarke
-- $1b/$5b /
-- $1d/$5d /
-- $1c/$5c /
-- $5e 
-- $3a/$3b : / ;
-- $3c/$3e < / >
 

charset = 
     " abcdefghijklmnopqrstuvwxyz^<" -- Pfeil oben Pfeil links
..[==[ !"#$%&'()*+,-./0123456789:;<=>?]==]
..   "?ABCDEFGHIJKLMNOPQRSTUVWXYZ"

function decodechar(iChar)
	return charset:sub(iChar+1,iChar+1) or "?"
end

-- 1-based offsets
function decodestring(str, iStart, iLen)
	assert(iStart+iLen-1 < str:len(), string.format("%d %d %d", str:len(), iStart, iLen))
	iStart = iStart or 1
	iLen = iLen or str:len()
	local chars = {}
	for i=iStart, iStart+iLen-1 do
		table.insert(chars, decodechar(str:byte(i)))
	end
	return table.concat(chars)
end



-- 0-14 Objektname, Screencodes aufgefllt mit Leerzeichen
-- 15 Bit 0 0: Objekt kann genommen werden, 1: unbewegliches Objekt
-- 15 Bit 1 0: Gegenstand, 1: Person

function getobj(objno)
	local bin, offset, obj
	if objno>=0 and objno<=0x37 then
		bin = game.oj
		offset = OJ_OBJ_OFFSET + objno*16
		obj = bin:sub(offset, offset+15)
	elseif objno>=0x40 and objno <= 0x7f then
		bin = game.o2
		offset = 0x03 + (objno-0x40)*16
		obj = bin:sub(offset, offset+15)
	else	
		return string.format(" **** Objekt $%02x **** ", objno), false, false, true
	end
	
	if obj then
		local fEmpty = false
		local strName = decodestring(obj, 1, 15)
		strName = strName:gsub("%s+$", "")
		if strName=="" then
			fEmpty = true
			strName = string.format("Objekt $%02x", objno)
		end
		
		local iAttr = obj:byte(16)
		local fTake = bit.band(iAttr, 1) == 0 
		local fPerson = bit.band(iAttr, 2) ~= 0
		assert(strName)
		return strName, fTake, fPerson, fEmpty
	end
end
	
function getflag(flagno)
	local offset
	local strName
	local fEmpty
	if flagno>=0 and flagno<=0x3f then
		if game.fl then
			offset = 3 + flagno*16
			strName = decodestring(game.fl, offset, 15)
			fEmpty = false
			strName = strName:gsub("%s+$", "")
			if strName=="" then
				strName = string.format("Flag %02x", flagno)
				fEmpty = true
			end
		else
			strName = string.format("Flag %02x", flagno)
			fEmpty = true
		end
		return strName, fEmpty
	end
end



function decodetextlines(strData, iOffset, iLines)
	local lines = {}
	local lastline = 0
	for i=1, iLines do
		local line = decodestring(strData, iOffset, 40)
		line = line:gsub("%s+$", "")
		lines[i] = line
		if line:len()>0 then
			lastline = i
		end
		iOffset = iOffset + 40		
	end
	return lines, lastline
end


function printtext(iNo, strData, iOffset)
	local lines, lastline = decodetextlines(strData, iOffset, 6)
	if lastline>0 then
		for i=1, lastline do
			printf("%x%x %s", i, iNo, lines[i])
		end	
		print()
	end
end

function printroomdescription(strUnpackedRoomData)
	local lines, lastline = decodetextlines(strUnpackedRoomData, 3, 10)
	for i=1, lastline do
		print(lines[i])
	end	
	print()
end

-- trennen an *, ^, ^^ 
function printroomdescription_hotspot(strUnpackedRoomData)
	local textlen = 400
	local strRaw = decodestring(strUnpackedRoomData, 3, textlen)
	strRaw = strRaw:gsub("%s+", " ")

		
	local lines = {}
	local iPos = 0 -- last word break
	local iLineStart = 1 -- starting position of line
	local iMaxlinelen = 64

	-- find the next space, *, ^, ^^ 
	-- str = word up to space, *, ...
	-- if it fits in the current line, append it
	-- otherwise, write out the current line and append to new line
	-- if str = *, ^, ^^, write out the current line and clear it
	
	while iPos < strRaw:len() do
		-- find the next break position
		local iStart, iEnd = strRaw:find("[%*%^ ]", iPos+1)
		local strBrk, strBrk2
		if iStart then
			strBrk = strRaw:sub(iStart, iEnd)
			strBr2 = strRaw:sub(iStart, iEnd+1)
			if strBrk2 == "^^" then
				iEnd = iEnd +1
			end
		else
			iEnd = strRaw:len()
			strBrk = ""
		end
		
		-- if break position - line start > max line len,
		-- break the line at the previous position
		if iEnd - iLineStart > iMaxlinelen then
			local strLine = strRaw:sub(iLineStart, iPos)
			table.insert(lines, strLine)
			iLineStart = iPos+1
		end

		-- if the break is not a space or the end of the text has 
		-- been reached, break the line here.
		if strBrk ~= " " or iEnd == strRaw:len() then
			local strLine = strRaw:sub(iLineStart, iEnd)
			table.insert(lines, strLine)
			iLineStart = iEnd+1
		end

		iPos = iEnd
	end
	
	-- if iLineStart < strRaw:len() then
	-- 	local strLine = strRaw:sub(iLineStart, strRaw:len())
	-- 	table.insert(lines, strLine)
	-- end
	
	for i, line in ipairs(lines) do
		print(line)
	end
	print()
end

function printroomdescription_hotspot__(strUnpackedRoomData)
	local textlen = 400
	local strRaw = decodestring(strUnpackedRoomData, 3, textlen)
	strRaw = strRaw:gsub("%s+", " ")

		
	local lines = {}
	local strLine = ""
	local iPos = 0 -- last word break
	local iMaxlinelen = 64

	-- find the next space, *, ^, ^^ 
	-- str = word up to space, *, ...
	-- if it fits in the current line, append it
	-- otherwise, write out the current line and append to new line
	-- if str = *, ^, ^^, write out the current line and clear it
	
	while iPos < strRaw:len() do
		-- find the next break position
		local iStart, iEnd = strRaw:find("[%*%^ ]", iPos+1)
		local strBrk, strBrk2, strWord
		if iStart then
			print(iStart, iEnd)
			strBrk = strRaw:sub(iStart, iEnd)
			strBr2 = strRaw:sub(iStart, iEnd+1)
			if strBrk2 == "^^" then
				iEnd = iEnd +1
			end
		else
			iEnd = strRaw:len()
			strBrk = ""
		end

		strWord = strRaw:sub(iPos+1, iEnd)

		-- if break position - line start > max line len,
		-- break the line at the previous position
		if strLine:len() + strWord:len() > iMaxlinelen then
			table.insert(lines, strLine)
			strLine = ""
		end

		-- append the word
		strLine = strLine .. strWord

		-- if the break is not a space or the end of the text has been reached, break the line here.
		if strBrk ~= " " or iEnd == strRaw:len() then
			table.insert(lines, strLine)
			strLine = ""
		end

		iPos = iEnd
	end
	
	for i, line in ipairs(lines) do
		print(line)
	end
	print()
end



-- Bedingungsblock
-- 0     Bit 0-5 Verb
-- 0     Bit 6-7 Vorgehensweise nach Ausfuhren der Aktionen
-- 1     Objektbedingung E oder Flagbedingung
-- 2     Objektbedingung B oder Flagbedingung
-- 3     Objektbedingung R oder Flagbedingung
-- 4-5   Aktion 1 mit Parameter
-- 6-7   Aktion 2 mit Parameter
-- 8-9   Aktion 3 mit Parameter
-- 10-11 Aktion 4 mit Parameter
-- 12-13 Aktion 5 mit Parameter
-- 14-15 Aktion 6 mit Parameter


-- $00-$7f Testen, ob Objekt im Befehl vorkommt
-- E: im Raum oder Inventar
-- B: im Besitz (Inventarobjekt)
-- R: im Raum
-- $80-$bf Test ob Flag gesetzt
-- $c0-$fe Test ob Flag gelscht
-- $ff Test uberspringen
-- 
-- 00 Aktionen ausfhren und Ende
-- 10 Aktionen ausfhren und weiter
-- 11 Aktionen ausfphren und weiter, aber den Test
-- der nchsten Bedingung berspringen


verbs={
[0]= "Bedingung ungltig",
[1] = "ffne",
[2] = "Verschiebe",
[3] = "Ein",
[4] = "Untersuche",
[5] = "Benutze",
[6] = "Kombiniere",
[7] = "Gib",
[8] = "Nimm",
[9] = "Verliere",
[10] = "Rede mit",
[11] = "Norden",
[12] = "Osten",
[13] = "Sden",
[14] = "Westen",
[63] = "Verb egal"
}

VERBNO_MASK = 0x3f
CONT_MASK = 0xc0
CONT_STRINGS = {[0] = "+", [0x40] = "??", [0x80] = "weiter", [0xc0] =  "weiter Aktionen"}


function act_msgno(iPar)
	local textno = iPar % 16
	local lineno = (iPar - textno)/16
	if lineno == 0 then
		return string.format("Text %d", textno)	
	else
		return string.format("Text %d Zeile %d", textno, lineno)
	end
end

function act_hexno(iPar)
	return string.format("%02x", iPar)
end

actions = {
[0] = 
{text = "----",                        paramfn = nil},
{text = "Gegenstand in Besitz",        paramfn = getobj},    -- Objektnr.
{text = "Lschen aus Besitz",          paramfn = getobj},    -- Objektnr.
{text = "Lschen aus Raum",            paramfn = getobj},    -- Objektnr.
{text = "Flag setzen",                 paramfn = getflag},   -- Flagname/Nr.
{text = "Flag lschen",                paramfn = getflag},   -- Flagname/Nr.
{text = "Meldung ausgeben",            paramfn = act_msgno}, -- Text x [Zeile y]
{text = "Allgemeinmeldung ausgeben",   paramfn = act_msgno}, -- Text x [Zeile y]
{text = "Sprung zu Raum",              paramfn = act_hexno}, -- Raumnr.
{text = "Game Over",                   paramfn = act_hexno}, -- Param.
{text = "Endsequenz",                  paramfn = nil},  
{text = "Auf Button warten",           paramfn = nil},
{text = "Gegenstand in Raum",          paramfn = getobj},    -- Ojektnr.   
}


function printaction(iNo, iCode, iPar)
	local strAct, strPar
	local act = actions[iCode]
	if act then
		strAct = act.text
		if act.paramfn then
			strPar = act.paramfn(iPar) or "???"
		end
	else
		strAct = "???"
	end
	strPar = strPar or ""
	printf("%02x %02x %-25s %s", iCode, iPar, strAct, strPar)
end

-- Flle:
-- ganze Bedingung/Aktionen sind 0 -> nur "leer" ausgeben
-- Verb = 0 -> gesperrt
-- Normale Bedingung

empty_cond_block = string.rep(string.char(0), 16)
ebrtab = {"E", "B", "R"}
function printcondition(iNo, strData, iOffset)
	if strData:sub(iOffset, iOffset+15) == empty_cond_block then
		--printf("Bedingung %02x: leer", iNo)
	else
		-- Bedingungen
		printf("Bedingung %02x", iNo)
		local iByte0, iByte1, iByte2, iByte3 = strData:byte(iOffset, iOffset + 3)
		local iVerbNo = bit.band(iByte0, VERBNO_MASK)
		local objconditions = {}
		local flagconditions = {}
		
		local strVerb
		if iVerbNo ~= 0x3f then
			strVerb = verbs[iVerbNo] or "???????????"
		else
			strVerb = ""
		end
			
		if iVerbNo ~= 0 then
			for i=1, 3 do
				local cond = strData:byte(iOffset+i)
				if cond <= 0x7f then
					local strObj = getobj(cond) or "????"
					table.insert(objconditions, string.format("%s:%s", ebrtab[i], strObj ))
				elseif cond <= 0xbf then
					table.insert(flagconditions, getflag(cond-0x80) .. " = 1")
				elseif cond < 0xff then
					table.insert(flagconditions, getflag(cond-0xc0) .. " = 0")
				end
			end
		end
		
		printf("%02x %02x %02x %02x %s %s %s", iByte0, iByte1, iByte2, iByte3,
			 strVerb, table.concat(objconditions, " "), table.concat(flagconditions, " "))

		-- Aktionen		
		iOffset = iOffset + 4
		for actno = 1, 6 do
			local iCode, iParam = strData:byte(iOffset, iOffset+1)
			if iCode ~= 0 then
				printaction(actno, iCode, iParam)
			end
			iOffset = iOffset + 2
		end	
		
		local iCont = bit.band(iByte0, CONT_MASK)
		if iCont ~= 0 then
			print(CONT_STRINGS[iCont])
		end

		print()
	
	end
	
end


function printobjmap()
	for roomno = 0, 127 do
		local objects = {}
		local lastslot = nil
		for slot = 0, 7 do
			local offset = 3 + 128*slot + roomno
			local entry = game.nw:byte(offset)
			local objno = bit.band(entry, 127)
			local sprite = bit.band(entry, 128)
			
			if objno ~= 0x7f then
				lastslot = slot
				objects[slot] = string.format("%02x %-15s %s", objno, getobj(objno), sprite == 0 and "Sprite" or " ")
			end
		
		end

		if lastslot then
			printf("Raum %02x:", roomno)
			for slot = 0, lastslot do
				if objects[slot] then
					printf("        %d %s", slot, objects[slot])
				end
			end
		else
			--printf("Raum %02x: keine Objekte", roomno)		
		end
	end
end

function printobjlist()
	for objno = 0, 0x7f do
		if objno < 0x38 or objno > 0x3f then
			local objname, fTake, fPerson, fEmpty = getobj(objno)
			if not fEmpty then
			
				printf("%02x %-15s %-13s %-10s", objno, objname, 
					fTake and "tragbar" or "nicht tragbar",
					fPerson and "Person" or "Gegenstand")
			end
		end
	end
end

function printflaglist()
	for i=0, 0x3f do
		local strName, fEmpty = getflag(i)
		if not fEmpty then 
			printf("%02x %s", i, strName)
		end
	end	
end


-- Raumfile
-- Adresse Inhalt
-- 8000-81df Raumbeschreibung (12*40 Zeichen)
-- 81e0 gesperrte Richtungen:
-- 1 - Norden, 2 - Osten, 4 - Sueden, 8 - Westen
-- 81e1 Musiknr.
-- 81e2-81e9 Sprites X-Tabelle
-- 81ea-81f1 Sprites Y-Tabelle
-- 8200-82ef Raumtext 1
-- ... ...
-- 8900-89ef Raumtext 8
-- 8a00-8bff 32 Bedingungen
-- 8c00-8ddf Videoram: 12x40 Bytes
-- 8e00-8fdf Colorram: 12x40 Bytes
-- 8fff Hintergrundfarbe
-- 9000-9eff Bitmap: 12x320 Bytes


-- Varianten bei Texten:
--PRINT_TEXTLIST = true
--TEXT_IN_LOGIC_ALWAYS = false
--TEXT_IN_LOGIC_FIRST = false

function printroom(iRoomNo, strRoomData)
	
	print()
	print( "=========")
	printf("Raum %02x", iRoomNo)
	print( "=========")
	print()
	
	-- Raumbeschreibung
	print("Raumbeschreibung: ")
	if not MODE_HOTSPOT then
		printroomdescription(strRoomData)
	else
		printroomdescription_hotspot(strRoomData)
	end
	print()
	
	-- Raumtexte
	print("Texte:")
	for i=0, 7 do
		local iOffset = 0x203 + i*256
		printtext(i, strRoomData, iOffset)
	end
	print()
	
	-- Bedingungen
	print("Bedingungen:")
	for i=0, 31 do
		local iOffset = 0xa03 + i*16
		printcondition(i, strRoomData, iOffset)
	end	
	print()
end

function printcommondata()
	print()
	print("==============")
	print("Allgemeindaten")
	print("==============")
	print()


	print("Objekte:")
	printobjlist()
	print()
	
	print("Gegenstandskarte:")
	printobjmap()
	print()
	
	print("Flags:")
	printflaglist()
	print()
	
	print("Allgemeintexte:")
	for i=0, 15 do
		printtext(i, game.ba, 0x0403 + i*256)
	end
	print()
	
	print("Allgemeinbedingungen:")
	for i=0, 63 do
		local iOffset = 3 + i*16
		printcondition(i, game.ba, iOffset)
	end
	print()
end

function loadgame(strDir)
	game = {}
	game.ba = loadFile(strDir.."/ba")
	game.oj = loadFile(strDir.."/oj")
	game.o2 = loadFile(strDir.."/o2")
	game.nw = loadFile(strDir.."/nw")
	game.fl = loadFile(strDir.."/fl")
	game.rooms = {}
	for iRoomNo = 0, 127 do
		local strFile = string.format("%02x", iRoomNo)
		local strPackedRoomData = loadFile(strDir .. "/" ..strFile)
		if strPackedRoomData then
			local strRoomData = depackD4(strPackedRoomData)
			game.rooms[iRoomNo] = strRoomData
		end
	end
end


-- 1 Norden, 2 Osten, 4 Sden,  8 Westen gesperrt
directionflags = {
[0] = 
"NOSW",
" OSW",
"N SW",
"  SW",
"NO W",
" O W",
"N  W",
"   W",
"NOS ",
" OS ",
"N S ",
"  S ",
"NO  ",
" O  ",
"N   ",
"    "}

function getroommatrixcell(iRoomNo, strRoomData)
	local iDir = strRoomData:byte(3 + 0x01e0)
	local iMus = strRoomData:byte(3 + 0x01e1)
	return {
		string.format(" %02x  ", iRoomNo),
		string.format (" %02x  ", iMus),
		directionflags[iDirFlags] .. " "
		}
end

function getroommatrixcell(iRoomNo, strRoomData)
	local iDir = strRoomData:byte(3 + 0x01e0)
	local iMus = strRoomData:byte(3 + 0x01e1)
	
	local strN = bit.band(iDir, 1)==0 and "+  +" or "+--+"
	local strO = bit.band(iDir, 2)==0 and " " or "|"
	local strS = bit.band(iDir, 4)==0 and "+  +" or "+--+"
	local strW = bit.band(iDir, 8)==0 and " " or "|"
	local strRoom = string.format("%02x", iRoomNo)
	local strMus = string.format("%02x", iMus)
	
	return {
		strN,
		strW .. strRoom .. strO,
		strW .. strMus .. strO,
		strS
		}
end


function printroommatrix()
	local ROWS = 4
	local COLS = 4
	local emptystring=string.rep(" ", COLS)
	
	for y = 0, 7 do
		local fLine = false
		local lines = {}
		for i=1, ROWS do lines[i] = {} end
		
		for x = 0, 15 do
			local iRoomNo = 16*y + x
			local strRoomData = game.rooms[iRoomNo]
			if strRoomData then
				local astr = getroommatrixcell(iRoomNo, strRoomData)
				for i=1, ROWS do table.insert(lines[i], astr[i])  end
				fLine = true			
			else
				for i=1, ROWS do table.insert(lines[i], emptystring)  end
			end
		end
		
		if fLine then
			for i=1, ROWS do print(table.concat(lines[i])) end
		else
			print()
		end
	end
end



if #arg ~= 1 then
    print (strUsage)
    os.exit(1)
else
    loadgame(arg[1])
    
    printcommondata()
    print("Rume:")
    printroommatrix()
    
    -- Rume ausgeben
    for iRoomNo = 0, 127 do
        if game.rooms[iRoomNo] then
            local strRoomData = game.rooms[iRoomNo]
            printroom(iRoomNo, strRoomData)
        end
    end
end
